@@ -161,7 +161,7 @@ showlimits(char *nam, int hard, int lim)
 	struct rlimit vals;
 	if (getrlimit(lim, &vals) < 0)
 	{
-	    zwarnnam(nam, "can't read limit: %e", NULL, errno);
+	    zwarnnam(nam, "can't read limit: %e", errno);
 	    return 1;
 	}
 	showlimitvalue(lim, hard ? vals.rlim_max : vals.rlim_cur);
@@ -199,7 +199,7 @@ printulimit(char *nam, int lim, int hard, int head)
 
 	if (getrlimit(lim, &vals) < 0)
 	{
-	    zwarnnam(nam, "can't read limit: %e", NULL, errno);
+	    zwarnnam(nam, "can't read limit: %e", errno);
 	    return 1;
 	}
 	limit = (hard) ? vals.rlim_max : vals.rlim_cur;
@@ -390,13 +390,13 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 	if (getrlimit(lim, &vals) < 0)
 	{
 	    /* best guess about error */
-	    zwarnnam(nam, "can't read limit: %e", NULL, errno);
+	    zwarnnam(nam, "can't read limit: %e", errno);
 	    return 1;
 	}
 	if (hard)
 	{
 	    if (val > vals.rlim_max && geteuid()) {
-		zwarnnam(nam, "can't raise hard limits", NULL, 0);
+		zwarnnam(nam, "can't raise hard limits");
 		return 1;
 	    }
 	    vals.rlim_max = val;
@@ -409,7 +409,7 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 	}
 	if (soft || !hard) {
 	    if (val > vals.rlim_max) {
-		zwarnnam(nam, "limit exceeds hard limit", NULL, 0);
+		zwarnnam(nam, "limit exceeds hard limit");
 		return 1;
 	    }
 	    else
@@ -419,12 +419,12 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 	{
 	    zwarnnam(nam,
 		     "warning: unrecognised limit %d, use -s to set",
-		     NULL, lim);
+		     lim);
 	    return 1;
 	}
 	else if (setrlimit(lim, &vals) < 0)
 	{
-	    zwarnnam(nam, "setrlimit failed: %e", NULL, errno);
+	    zwarnnam(nam, "setrlimit failed: %e", errno);
 	    return 1;
 	}
     } else {
@@ -433,7 +433,7 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 	if (hard) {
 	    /* can only raise hard limits if running as root */
 	    if (val > current_limits[lim].rlim_max && geteuid()) {
-		zwarnnam(nam, "can't raise hard limits", NULL, 0);
+		zwarnnam(nam, "can't raise hard limits");
 		return 1;
 	    } else {
 		limits[lim].rlim_max = val;
@@ -448,13 +448,13 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 		{
 		    /* ulimit does this */
 		    if (val > current_limits[lim].rlim_max && geteuid()) {
-			zwarnnam(nam, "value exceeds hard limit", NULL, 0);
+			zwarnnam(nam, "value exceeds hard limit");
 			return 1;
 		    }
 		    limits[lim].rlim_max = limits[lim].rlim_cur = val;
 		} else {
 		    /* but limit does this */
-		    zwarnnam(nam, "limit exceeds hard limit", NULL, 0);
+		    zwarnnam(nam, "limit exceeds hard limit");
 		    return 1;
 		}
 	    } else
@@ -505,7 +505,7 @@ bin_limit(char *nam, char **argv, Options ops, UNUSED(int func))
 	if (lim < 0) {
 	    zwarnnam(nam,
 		     (lim == -2) ? "ambiguous resource specification: %s"
-		     : "no such resource: %s", s, 0);
+		     : "no such resource: %s", s);
 	    return 1;
 	}
 	/* without value for limit, display the current limit */
@@ -517,7 +517,7 @@ bin_limit(char *nam, char **argv, Options ops, UNUSED(int func))
 	    if (*s)
 	    {
 		/* unknown limit, no idea how to scale */
-		zwarnnam(nam, "unknown scaling factor: %s", s, 0);
+		zwarnnam(nam, "unknown scaling factor: %s", s);
 		return 1;
 	    }
 	}
@@ -535,7 +535,7 @@ bin_limit(char *nam, char **argv, Options ops, UNUSED(int func))
 		else if (*s == ':')
 		    val = val * 60 + zstrtorlimt(s + 1, &s, 10);
 		else {
-		    zwarnnam(nam, "unknown scaling factor: %s", s, 0);
+		    zwarnnam(nam, "unknown scaling factor: %s", s);
 		    return 1;
 		}
 	    }
@@ -545,7 +545,7 @@ bin_limit(char *nam, char **argv, Options ops, UNUSED(int func))
 	    char *t = s;
 	    val = zstrtorlimt(t, &s, 10);
 	    if (s == t) {
-		zwarnnam(nam, "limit must be a number", NULL, 0);
+		zwarnnam(nam, "limit must be a number");
 		return 1;
 	    }
 	} else {
@@ -558,7 +558,7 @@ bin_limit(char *nam, char **argv, Options ops, UNUSED(int func))
 	    } else if ((*s == 'M' || *s == 'm') && !s[1])
 		val *= 1024L * 1024;
 	    else {
-		zwarnnam(nam, "unknown scaling factor: %s", s, 0);
+		zwarnnam(nam, "unknown scaling factor: %s", s);
 		return 1;
 	    }
 	}
@@ -577,12 +577,12 @@ do_unlimit(char *nam, int lim, int hard, int soft, int set, int euid)
 	struct rlimit vals;
 	if (getrlimit(lim, &vals) < 0)
 	{
-	    zwarnnam(nam, "can't read limit: %e", NULL, errno);
+	    zwarnnam(nam, "can't read limit: %e", errno);
 	    return 1;
 	}
 	if (hard) {
 	    if (euid && vals.rlim_max != RLIM_INFINITY) {
-		zwarnnam(nam, "can't remove hard limits", NULL, 0);
+		zwarnnam(nam, "can't remove hard limits");
 		return 1;
 	    } else
 		vals.rlim_max = RLIM_INFINITY;
@@ -591,17 +591,16 @@ do_unlimit(char *nam, int lim, int hard, int soft, int set, int euid)
 	    vals.rlim_cur = vals.rlim_max;
 	if (!set) {
 	    zwarnnam(nam,
-		     "warning: unrecognised limit %d, use -s to set",
-		     NULL, lim);
+		     "warning: unrecognised limit %d, use -s to set", lim);
 	    return 1;
 	} else if (setrlimit(lim, &vals) < 0) {
-	    zwarnnam(nam, "setrlimit failed: %e", NULL, errno);
+	    zwarnnam(nam, "setrlimit failed: %e", errno);
 	    return 1;
 	}
     } else {
 	if (hard) {
 	    if (euid && current_limits[lim].rlim_max != RLIM_INFINITY) {
-		zwarnnam(nam, "can't remove hard limits", NULL, 0);
+		zwarnnam(nam, "can't remove hard limits");
 		return 1;
 	    } else
 		limits[lim].rlim_max = RLIM_INFINITY;
@@ -640,7 +639,7 @@ bin_unlimit(char *nam, char **argv, Options ops, UNUSED(int func))
 	if (OPT_ISSET(ops,'s'))
 	    ret += setlimits(nam);
 	if (ret)
-	    zwarnnam(nam, "can't remove hard limits", NULL, 0);
+	    zwarnnam(nam, "can't remove hard limits");
     } else {
 	for (; *argv; argv++) {
 	    /* Search for the appropriate resource name.  When a name     *
@@ -663,7 +662,7 @@ bin_unlimit(char *nam, char **argv, Options ops, UNUSED(int func))
 	    if (lim < 0) {
 		zwarnnam(nam,
 			 (lim == -2) ? "ambiguous resource specification: %s"
-			 : "no such resource: %s", *argv, 0);
+			 : "no such resource: %s", *argv);
 		return 1;
 	    }
 	    else if (do_unlimit(nam, lim, hard, !hard, OPT_ISSET(ops, 's'),
@@ -686,7 +685,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
     do {
 	options = *argv;
 	if (options && *options == '-' && !options[1]) {
-	    zwarnnam(name, "missing option letter", NULL, 0);
+	    zwarnnam(name, "missing option letter");
 	    return 1;
 	}
 	res = -1;
@@ -709,8 +708,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 		    } else if (*argv) {
 			res = (int)zstrtol(*argv++, NULL, 10);
 		    } else {
-			zwarnnam(name, "number required after -N",
-				 NULL, 0);
+			zwarnnam(name, "number required after -N");
 			return 1;
 		    }
 		    /*
@@ -721,8 +719,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 		    break;
 		case 'a':
 		    if (resmask) {
-			zwarnnam(name, "no limits allowed with -a",
-				 NULL, 0);
+			zwarnnam(name, "no limits allowed with -a");
 			return 1;
 		    }
 		    all = 1;
@@ -800,7 +797,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 # endif
 		default:
 		    /* unrecognised limit */
-		    zwarnnam(name, "bad option: -%c", NULL, *options);
+		    zwarnnam(name, "bad option: -%c", *options);
 		    return 1;
 		}
 		if (options[1]) {
@@ -808,8 +805,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 		    nres++;
 		}
 		if (all && res != -1) {
-		    zwarnnam(name, "no limits allowed with -a",
-			     NULL, 0);
+		    zwarnnam(name, "no limits allowed with -a");
 		    return 1;
 		}
 	    }
@@ -826,7 +822,7 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 	    continue;
 	}
 	if (all) {
-	    zwarnnam(name, "no arguments allowed after -a", NULL, 0);
+	    zwarnnam(name, "no arguments allowed after -a");
 	    return 1;
 	}
 	if (res < 0)
@@ -893,6 +889,14 @@ static struct builtin bintab[] = {
     BUILTIN("unlimit", 0, bin_unlimit, 0, -1, 0, "hs", NULL),
 };
 
+static struct features module_features = {
+    bintab, sizeof(bintab)/sizeof(*bintab),
+    NULL, 0,
+    NULL, 0,
+    NULL, 0,
+    0
+};
+
 /**/
 int
 setup_(UNUSED(Module m))
@@ -900,18 +904,33 @@ setup_(UNUSED(Module m))
     return 0;
 }
 
+/**/
+int
+features_(Module m, char ***features)
+{
+    *features = featuresarray(m, &module_features);
+    return 0;
+}
+
+/**/
+int
+enables_(Module m, int **enables)
+{
+    return handlefeatures(m, &module_features, enables);
+}
+
 /**/
 int
 boot_(Module m)
 {
-    return !addbuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
+    return 0;
 }
 
 /**/
 int
 cleanup_(Module m)
 {
-    deletebuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
+    return setfeatureenables(m, &module_features, NULL);
     return 0;
 }
 
